@@ -63,13 +63,6 @@ pr_debug(DRIVER_NAME " [%s()]: " f, __func__, ## x)
 #define TIMEOUT_VAL 0xE
 #define BCM2835_SDHCI_WRITE_DELAY(f)	(((2 * 1000000) / f) + 1)
 
-#ifndef BCM2708_PERI_BASE
- #define BCM2708_PERI_BASE 0x20000000
-#endif
-
-/* FIXME: Needs IOMMU support */
-#define BCM2835_VCMMU_SHIFT		(0x7E000000 - BCM2708_PERI_BASE)
-
 
 unsigned mmc_debug;
 unsigned mmc_debug2;
@@ -78,7 +71,7 @@ struct bcm2835_host {
 	spinlock_t				lock;
 
 	void __iomem			*ioaddr;
-	u32						phys_addr;
+	u32						bus_addr;
 
 	struct mmc_host			*mmc;
 
@@ -690,12 +683,10 @@ void bcm2835_mmc_send_command(struct bcm2835_host *host, struct mmc_command *cmd
 	}
 
 	timeout = jiffies;
-#ifdef CONFIG_ARCH_BCM2835
 	if (!cmd->data && cmd->busy_timeout > 9000)
 		timeout += DIV_ROUND_UP(cmd->busy_timeout, 1000) * HZ + HZ;
 	else
-#endif
-	timeout += 10 * HZ;
+		timeout += 10 * HZ;
 	mod_timer(&host->timer, timeout);
 
 	host->cmd = cmd;
@@ -990,9 +981,6 @@ static irqreturn_t bcm2835_mmc_irq(int irq, void *dev_id)
 	struct bcm2835_host *host = dev_id;
 	u32 intmask, mask, unexpected = 0;
 	int max_loops = 16;
-#ifndef CONFIG_ARCH_BCM2835
-	int cardint = 0;
-#endif
 
 	spin_lock(&host->lock);
 
@@ -1021,13 +1009,9 @@ static irqreturn_t bcm2835_mmc_irq(int irq, void *dev_id)
 				mmc_hostname(host->mmc));
 
 		if (intmask & SDHCI_INT_CARD_INT) {
-#ifndef CONFIG_ARCH_BCM2835
-			cardint = 1;
-#else
 			bcm2835_mmc_enable_sdio_irq_nolock(host, false);
 			host->thread_isr |= SDHCI_INT_CARD_INT;
 			result = IRQ_WAKE_THREAD;
-#endif
 		}
 
 		intmask &= ~(SDHCI_INT_CARD_INSERT | SDHCI_INT_CARD_REMOVE |
@@ -1054,15 +1038,9 @@ out:
 		bcm2835_mmc_dumpregs(host);
 	}
 
-#ifndef CONFIG_ARCH_BCM2835
-	if (cardint)
-		mmc_signal_sdio_irq(host->mmc);
-#endif
-
 	return result;
 }
 
-#ifdef CONFIG_ARCH_BCM2835
 static irqreturn_t bcm2835_mmc_thread_irq(int irq, void *dev_id)
 {
 	struct bcm2835_host *host = dev_id;
@@ -1085,7 +1063,6 @@ static irqreturn_t bcm2835_mmc_thread_irq(int irq, void *dev_id)
 
 	return isr ? IRQ_HANDLED : IRQ_NONE;
 }
-#endif
 
 
 
@@ -1329,14 +1306,15 @@ static int bcm2835_mmc_add_host(struct bcm2835_host *host)
 
 	/* SDHCI_QUIRK_DATA_TIMEOUT_USES_SDCLK */
 	host->timeout_clk = mmc->f_max / 1000;
-#ifdef CONFIG_ARCH_BCM2835
 	mmc->max_busy_timeout = (1 << 27) / host->timeout_clk;
-#endif
+
 	/* host controller capabilities */
 	mmc->caps |= MMC_CAP_CMD23 | MMC_CAP_ERASE | MMC_CAP_NEEDS_POLL |
 		MMC_CAP_SDIO_IRQ | MMC_CAP_SD_HIGHSPEED |
 		MMC_CAP_MMC_HIGHSPEED;
 
+	mmc->caps2 |= MMC_CAP2_SDIO_IRQ_NOTHREAD;
+
 	host->flags = SDHCI_AUTO_CMD23;
 
 	dev_info(dev, "mmc_debug:%x mmc_debug2:%x\n", mmc_debug, mmc_debug2);
@@ -1359,7 +1337,7 @@ static int bcm2835_mmc_add_host(struct bcm2835_host *host)
 
 		cfg.direction = DMA_MEM_TO_DEV;
 		cfg.src_addr = 0;
-		cfg.dst_addr = host->phys_addr + SDHCI_BUFFER;
+		cfg.dst_addr = host->bus_addr + SDHCI_BUFFER;
 
 		ret = dmaengine_slave_config(host->dma_chan_rxtx, &cfg);
 
@@ -1367,7 +1345,7 @@ static int bcm2835_mmc_add_host(struct bcm2835_host *host)
 			host->dma_cfg_tx = cfg;
 
 			cfg.direction = DMA_DEV_TO_MEM;
-			cfg.src_addr = host->phys_addr + SDHCI_BUFFER;
+			cfg.src_addr = host->bus_addr + SDHCI_BUFFER;
 			cfg.dst_addr = 0;
 
 			ret = dmaengine_slave_config(host->dma_chan_rxtx, &cfg);
@@ -1403,14 +1381,9 @@ static int bcm2835_mmc_add_host(struct bcm2835_host *host)
 	init_waitqueue_head(&host->buf_ready_int);
 
 	bcm2835_mmc_init(host, 0);
-#ifndef CONFIG_ARCH_BCM2835
-	ret = devm_request_irq(dev, host->irq, bcm2835_mmc_irq, 0,
-			       mmc_hostname(mmc), host);
-#else
 	ret = devm_request_threaded_irq(dev, host->irq, bcm2835_mmc_irq,
 					bcm2835_mmc_thread_irq, IRQF_SHARED,
 					mmc_hostname(mmc), host);
-#endif
 	if (ret) {
 		dev_err(dev, "Failed to request IRQ %d: %d\n", host->irq, ret);
 		goto untasklet;
@@ -1435,6 +1408,7 @@ static int bcm2835_mmc_probe(struct platform_device *pdev)
 	struct resource *iomem;
 	struct bcm2835_host *host;
 	struct mmc_host *mmc;
+	const __be32 *addr;
 	int ret;
 
 	mmc = mmc_alloc_host(sizeof(*host), dev);
@@ -1454,7 +1428,16 @@ static int bcm2835_mmc_probe(struct platform_device *pdev)
 		goto err;
 	}
 
-	host->phys_addr = iomem->start + BCM2835_VCMMU_SHIFT;
+	addr = of_get_address(node, 0, NULL, NULL);
+	if (!addr) {
+		dev_err(dev, "could not get DMA-register address\n");
+		return -ENODEV;
+	}
+	host->bus_addr = be32_to_cpup(addr);
+	pr_debug(" - ioaddr %lx, iomem->start %lx, bus_addr %lx\n",
+		 (unsigned long)host->ioaddr,
+		 (unsigned long)iomem->start,
+		 (unsigned long)host->bus_addr);
 
 #ifndef FORCE_PIO
 	if (node) {
