@@ -38,6 +38,7 @@
 #include <linux/mmc/mmc.h>
 #include <linux/mmc/host.h>
 #include <linux/mmc/sd.h>
+#include <linux/mmc/sdio.h>
 #include <linux/scatterlist.h>
 #include <linux/of_address.h>
 #include <linux/of_irq.h>
@@ -50,6 +51,7 @@
 #include <linux/of_dma.h>
 #include <linux/time.h>
 #include <linux/workqueue.h>
+#include <soc/bcm2835/raspberrypi-firmware.h>
 
 #define DRIVER_NAME "sdhost-bcm2835"
 
@@ -136,19 +138,12 @@
 
 #define MHZ 1000000
 
-#ifndef BCM2708_PERI_BASE
- #define BCM2708_PERI_BASE 0x20000000
-#endif
-
-/* FIXME: Needs IOMMU support */
-#define BCM2835_VCMMU_SHIFT		(0x7E000000 - BCM2708_PERI_BASE)
-
 
 struct bcm2835_host {
 	spinlock_t		lock;
 
 	void __iomem		*ioaddr;
-	u32			phys_addr;
+	u32			bus_addr;
 
 	struct mmc_host		*mmc;
 
@@ -190,6 +185,8 @@ struct bcm2835_host {
 	unsigned int			use_sbc:1;		/* Send CMD23 */
 
 	unsigned int			debug:1;		/* Enable debug output */
+	unsigned int			firmware_sets_cdiv:1;	/* Let the firmware manage the clock */
+	unsigned int			reset_clock:1;		/* Reset the clock fore the next request */
 
 	/*DMA part*/
 	struct dma_chan			*dma_chan_rxtx;		/* DMA channel for reads and writes */
@@ -210,6 +207,7 @@ struct bcm2835_host {
 	struct timeval			stop_time;	/* when the last stop was issued */
 	u32				delay_after_stop; /* minimum time between stop and subsequent data transfer */
 	u32				delay_after_this_stop; /* minimum time between this stop and subsequent data transfer */
+	u32				user_overclock_50; /* User's preferred frequency to use when 50MHz is requested (in MHz) */
 	u32				overclock_50;	/* frequency to use when 50MHz is requested (in MHz) */
 	u32				overclock;	/* Current frequency if overclocked, else zero */
 	u32				pio_limit;	/* Maximum block count for PIO (0 = always DMA) */
@@ -237,7 +235,7 @@ static void __iomem *timer_base;
 #define LOG_ENTRIES (256*1)
 #define LOG_SIZE (sizeof(LOG_ENTRY_T)*LOG_ENTRIES)
 
-static void log_init(void)
+static void log_init(u32 bus_to_phys)
 {
 	spin_lock_init(&log_lock);
 	sdhost_log_buf = dma_zalloc_coherent(NULL, LOG_SIZE, &sdhost_log_addr,
@@ -245,7 +243,7 @@ static void log_init(void)
 	if (sdhost_log_buf) {
 		pr_info("sdhost: log_buf @ %p (%x)\n",
 			sdhost_log_buf, sdhost_log_addr);
-		timer_base = ioremap_nocache(BCM2708_PERI_BASE + 0x3000, SZ_4K);
+		timer_base = ioremap_nocache(bus_to_phys + 0x7e003000, SZ_4K);
 		if (!timer_base)
 			pr_err("sdhost: failed to remap timer\n");
 	}
@@ -286,7 +284,7 @@ static void log_dump(void)
 		do {
 			entry = sdhost_log_buf + idx;
 			if (entry->event[0] != '\0')
-				pr_err("[%08x] %.4s %x %x\n",
+				pr_info("[%08x] %.4s %x %x\n",
 				       entry->timestamp,
 				       entry->event,
 				       entry->param1,
@@ -302,7 +300,7 @@ static void log_dump(void)
 
 #else
 
-#define log_init() (void)0
+#define log_init(x) (void)0
 #define log_event(event, param1, param2) (void)0
 #define log_dump() (void)0
 
@@ -328,7 +326,7 @@ static void bcm2835_sdhost_dumpcmd(struct bcm2835_host *host,
 				   const char *label)
 {
 	if (cmd)
-		pr_err("%s:%c%s op %d arg 0x%x flags 0x%x - resp %08x %08x %08x %08x, err %d\n",
+		pr_info("%s:%c%s op %d arg 0x%x flags 0x%x - resp %08x %08x %08x %08x, err %d\n",
 			mmc_hostname(host->mmc),
 			(cmd == host->cmd) ? '>' : ' ',
 			label, cmd->opcode, cmd->arg, cmd->flags,
@@ -343,7 +341,7 @@ static void bcm2835_sdhost_dumpregs(struct bcm2835_host *host)
 		bcm2835_sdhost_dumpcmd(host, host->mrq->sbc, "sbc");
 		bcm2835_sdhost_dumpcmd(host, host->mrq->cmd, "cmd");
 		if (host->mrq->data)
-			pr_err("%s: data blocks %x blksz %x - err %d\n",
+			pr_info("%s: data blocks %x blksz %x - err %d\n",
 			       mmc_hostname(host->mmc),
 			       host->mrq->data->blocks,
 			       host->mrq->data->blksz,
@@ -351,53 +349,53 @@ static void bcm2835_sdhost_dumpregs(struct bcm2835_host *host)
 		bcm2835_sdhost_dumpcmd(host, host->mrq->stop, "stop");
 	}
 
-	pr_err("%s: =========== REGISTER DUMP ===========\n",
+	pr_info("%s: =========== REGISTER DUMP ===========\n",
 		mmc_hostname(host->mmc));
 
-	pr_err("%s: SDCMD  0x%08x\n",
+	pr_info("%s: SDCMD  0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDCMD));
-	pr_err("%s: SDARG  0x%08x\n",
+	pr_info("%s: SDARG  0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDARG));
-	pr_err("%s: SDTOUT 0x%08x\n",
+	pr_info("%s: SDTOUT 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDTOUT));
-	pr_err("%s: SDCDIV 0x%08x\n",
+	pr_info("%s: SDCDIV 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDCDIV));
-	pr_err("%s: SDRSP0 0x%08x\n",
+	pr_info("%s: SDRSP0 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDRSP0));
-	pr_err("%s: SDRSP1 0x%08x\n",
+	pr_info("%s: SDRSP1 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDRSP1));
-	pr_err("%s: SDRSP2 0x%08x\n",
+	pr_info("%s: SDRSP2 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDRSP2));
-	pr_err("%s: SDRSP3 0x%08x\n",
+	pr_info("%s: SDRSP3 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDRSP3));
-	pr_err("%s: SDHSTS 0x%08x\n",
+	pr_info("%s: SDHSTS 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDHSTS));
-	pr_err("%s: SDVDD  0x%08x\n",
+	pr_info("%s: SDVDD  0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDVDD));
-	pr_err("%s: SDEDM  0x%08x\n",
+	pr_info("%s: SDEDM  0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDEDM));
-	pr_err("%s: SDHCFG 0x%08x\n",
+	pr_info("%s: SDHCFG 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDHCFG));
-	pr_err("%s: SDHBCT 0x%08x\n",
+	pr_info("%s: SDHBCT 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDHBCT));
-	pr_err("%s: SDHBLC 0x%08x\n",
+	pr_info("%s: SDHBLC 0x%08x\n",
 		mmc_hostname(host->mmc),
 		bcm2835_sdhost_read(host, SDHBLC));
 
-	pr_err("%s: ===========================================\n",
+	pr_info("%s: ===========================================\n",
 		mmc_hostname(host->mmc));
 }
 
@@ -437,7 +435,7 @@ static void bcm2835_sdhost_reset_internal(struct bcm2835_host *host)
 	host->clock = 0;
 	host->sectors = 0;
 	bcm2835_sdhost_write(host, host->hcfg, SDHCFG);
-	bcm2835_sdhost_write(host, host->cdiv, SDCDIV);
+	bcm2835_sdhost_write(host, SDCDIV_MAX_CDIV, SDCDIV);
 	mmiowb();
 }
 
@@ -612,7 +610,7 @@ static void bcm2835_sdhost_read_block_pio(struct bcm2835_host *host)
 				    (fsm_state != SDEDM_FSM_READCRC)) {
 					hsts = bcm2835_sdhost_read(host,
 								   SDHSTS);
-					pr_err("%s: fsm %x, hsts %x\n",
+					pr_info("%s: fsm %x, hsts %x\n",
 					       mmc_hostname(host->mmc),
 					       fsm_state, hsts);
 					if (hsts & SDHSTS_ERROR_MASK)
@@ -702,7 +700,7 @@ static void bcm2835_sdhost_write_block_pio(struct bcm2835_host *host)
 				    (fsm_state != SDEDM_FSM_WRITESTART2)) {
 					hsts = bcm2835_sdhost_read(host,
 								   SDHSTS);
-					pr_err("%s: fsm %x, hsts %x\n",
+					pr_info("%s: fsm %x, hsts %x\n",
 					       mmc_hostname(host->mmc),
 					       fsm_state, hsts);
 					if (hsts & SDHSTS_ERROR_MASK)
@@ -957,9 +955,10 @@ bool bcm2835_sdhost_send_command(struct bcm2835_host *host,
 
 	while (bcm2835_sdhost_read(host, SDCMD) & SDCMD_NEW_FLAG) {
 		if (timeout == 0) {
-			pr_err("%s: previous command never completed.\n",
+			pr_warn("%s: previous command never completed.\n",
 				mmc_hostname(host->mmc));
-			bcm2835_sdhost_dumpregs(host);
+			if (host->debug)
+				bcm2835_sdhost_dumpregs(host);
 			cmd->error = -EILSEQ;
 			tasklet_schedule(&host->finish_tasklet);
 			return false;
@@ -1184,9 +1183,8 @@ static void bcm2835_sdhost_finish_command(struct bcm2835_host *host,
 		retries = 1; // We've already waited long enough this time
 	}
 
-	retries = host->cmd_quick_poll_retries;
 	for (sdcmd = bcm2835_sdhost_read(host, SDCMD);
-	     (sdcmd & SDCMD_NEW_FLAG) && !(sdcmd & SDCMD_FAIL_FLAG) && retries;
+	     (sdcmd & SDCMD_NEW_FLAG) && retries;
 	     retries--) {
 		cpu_relax();
 		sdcmd = bcm2835_sdhost_read(host, SDCMD);
@@ -1209,18 +1207,19 @@ static void bcm2835_sdhost_finish_command(struct bcm2835_host *host,
 			usleep_range(1, 10);
 			spin_lock_irqsave(&host->lock, *irq_flags);
 			sdcmd = bcm2835_sdhost_read(host, SDCMD);
-			if (!(sdcmd & SDCMD_NEW_FLAG) ||
-			    (sdcmd & SDCMD_FAIL_FLAG))
+			if (!(sdcmd & SDCMD_NEW_FLAG))
 				break;
 		}
 	}
 
 	/* Check for errors */
 	if (sdcmd & SDCMD_NEW_FLAG) {
-		pr_err("%s: command never completed.\n",
-		       mmc_hostname(host->mmc));
-		bcm2835_sdhost_dumpregs(host);
-		host->cmd->error = -EIO;
+		if (host->debug) {
+			pr_err("%s: command %d never completed.\n",
+			       mmc_hostname(host->mmc), host->cmd->opcode);
+			bcm2835_sdhost_dumpregs(host);
+		}
+		host->cmd->error = -EILSEQ;
 		tasklet_schedule(&host->finish_tasklet);
 		return;
 	} else if (sdcmd & SDCMD_FAIL_FLAG) {
@@ -1242,15 +1241,14 @@ static void bcm2835_sdhost_finish_command(struct bcm2835_host *host,
 		} else {
 			if (sdhsts & SDHSTS_CMD_TIME_OUT) {
 				if (host->debug)
-					pr_err("%s: command %d timeout\n",
+					pr_warn("%s: command %d timeout\n",
 					       mmc_hostname(host->mmc),
 					       host->cmd->opcode);
 				host->cmd->error = -ETIMEDOUT;
 			} else {
-				pr_err("%s: unexpected command %d error\n",
+				pr_warn("%s: unexpected command %d error\n",
 				       mmc_hostname(host->mmc),
 				       host->cmd->opcode);
-				bcm2835_sdhost_dumpregs(host);
 				host->cmd->error = -EILSEQ;
 			}
 			tasklet_schedule(&host->finish_tasklet);
@@ -1374,8 +1372,10 @@ static void bcm2835_sdhost_busy_irq(struct bcm2835_host *host, u32 intmask)
 		} else if (intmask & SDHSTS_CMD_TIME_OUT)
 			host->cmd->error = -ETIMEDOUT;
 
-		log_dump();
-		bcm2835_sdhost_dumpregs(host);
+		if (host->debug) {
+			log_dump();
+			bcm2835_sdhost_dumpregs(host);
+		}
 	}
 	else
 		bcm2835_sdhost_finish_command(host, NULL);
@@ -1510,6 +1510,7 @@ void bcm2835_sdhost_set_clock(struct bcm2835_host *host, unsigned int clock)
 {
 	int div = 0; /* Initialized for compiler warning */
 	unsigned int input_clock = clock;
+	unsigned long flags;
 
 	if (host->debug)
 		pr_info("%s: set_clock(%d)\n", mmc_hostname(host->mmc), clock);
@@ -1541,62 +1542,89 @@ void bcm2835_sdhost_set_clock(struct bcm2835_host *host, unsigned int clock)
 
 	host->mmc->actual_clock = 0;
 
-	if (clock < 100000) {
-	    /* Can't stop the clock, but make it as slow as possible
-	     * to show willing
-	     */
-	    host->cdiv = SDCDIV_MAX_CDIV;
-	    bcm2835_sdhost_write(host, host->cdiv, SDCDIV);
-	    return;
+	if (host->firmware_sets_cdiv) {
+		u32 msg[3] = { clock, 0, 0 };
+
+		rpi_firmware_property(rpi_firmware_get(NULL),
+				      RPI_FIRMWARE_SET_SDHOST_CLOCK,
+				      &msg, sizeof(msg));
+
+		clock = max(msg[1], msg[2]);
+		spin_lock_irqsave(&host->lock, flags);
+	} else {
+		spin_lock_irqsave(&host->lock, flags);
+		if (clock < 100000) {
+			/* Can't stop the clock, but make it as slow as
+			 * possible to show willing
+			 */
+			host->cdiv = SDCDIV_MAX_CDIV;
+			bcm2835_sdhost_write(host, host->cdiv, SDCDIV);
+			mmiowb();
+			spin_unlock_irqrestore(&host->lock, flags);
+			return;
+		}
+
+		div = host->max_clk / clock;
+		if (div < 2)
+			div = 2;
+		if ((host->max_clk / div) > clock)
+			div++;
+		div -= 2;
+
+		if (div > SDCDIV_MAX_CDIV)
+			div = SDCDIV_MAX_CDIV;
+
+		clock = host->max_clk / (div + 2);
+
+		host->cdiv = div;
+		bcm2835_sdhost_write(host, host->cdiv, SDCDIV);
+
+		if (host->debug)
+			pr_info("%s: clock=%d -> max_clk=%d, cdiv=%x "
+				"(actual clock %d)\n",
+				mmc_hostname(host->mmc), input_clock,
+				host->max_clk, host->cdiv,
+				clock);
 	}
 
-	div = host->max_clk / clock;
-	if (div < 2)
-		div = 2;
-	if ((host->max_clk / div) > clock)
-		div++;
-	div -= 2;
-
-	if (div > SDCDIV_MAX_CDIV)
-	    div = SDCDIV_MAX_CDIV;
-
-	clock = host->max_clk / (div + 2);
-	host->mmc->actual_clock = clock;
-
 	/* Calibrate some delays */
 
 	host->ns_per_fifo_word = (1000000000/clock) *
 		((host->mmc->caps & MMC_CAP_4_BIT_DATA) ? 8 : 32);
 
-	if (clock > input_clock) {
-		/* Save the closest value, to make it easier
-		   to reduce in the event of error */
-		host->overclock_50 = (clock/MHZ);
+	if (input_clock == 50 * MHZ) {
+		if (clock > input_clock) {
+			/* Save the closest value, to make it easier
+			   to reduce in the event of error */
+			host->overclock_50 = (clock/MHZ);
 
-		if (clock != host->overclock) {
-			pr_warn("%s: overclocking to %dHz\n",
-				mmc_hostname(host->mmc), clock);
-			host->overclock = clock;
+			if (clock != host->overclock) {
+				pr_info("%s: overclocking to %dHz\n",
+					mmc_hostname(host->mmc), clock);
+				host->overclock = clock;
+			}
+		} else if (host->overclock) {
+			host->overclock = 0;
+			if (clock == 50 * MHZ)
+				pr_warn("%s: cancelling overclock\n",
+					mmc_hostname(host->mmc));
 		}
-	}
-	else if (host->overclock)
-	{
+	} else if (input_clock == 0) {
+		/* Reset the preferred overclock when the clock is stopped.
+		 * This always happens during initialisation. */
+		host->overclock_50 = host->user_overclock_50;
 		host->overclock = 0;
-		if (clock == 50 * MHZ)
-			pr_warn("%s: cancelling overclock\n",
-				mmc_hostname(host->mmc));
 	}
 
-	host->cdiv = div;
-	bcm2835_sdhost_write(host, host->cdiv, SDCDIV);
-
 	/* Set the timeout to 500ms */
-	bcm2835_sdhost_write(host, host->mmc->actual_clock/2, SDTOUT);
+	bcm2835_sdhost_write(host, clock/2, SDTOUT);
 
-	if (host->debug)
-		pr_info("%s: clock=%d -> max_clk=%d, cdiv=%x (actual clock %d)\n",
-			mmc_hostname(host->mmc), input_clock,
-			host->max_clk, host->cdiv, host->mmc->actual_clock);
+	host->mmc->actual_clock = clock;
+	host->clock = input_clock;
+	host->reset_clock = 0;
+
+	mmiowb();
+	spin_unlock_irqrestore(&host->lock, flags);
 }
 
 static void bcm2835_sdhost_request(struct mmc_host *mmc, struct mmc_request *mrq)
@@ -1645,6 +1673,9 @@ static void bcm2835_sdhost_request(struct mmc_host *mmc, struct mmc_request *mrq
 	    (mrq->data->blocks > host->pio_limit))
 		bcm2835_sdhost_prepare_dma(host, mrq->data);
 
+	if (host->reset_clock)
+	    bcm2835_sdhost_set_clock(host, host->clock);
+
 	spin_lock_irqsave(&host->lock, flags);
 
 	WARN_ON(host->mrq != NULL);
@@ -1656,13 +1687,15 @@ static void bcm2835_sdhost_request(struct mmc_host *mmc, struct mmc_request *mrq
 	log_event("REQ<", (u32)mrq, edm);
 	if ((fsm != SDEDM_FSM_IDENTMODE) &&
 	    (fsm != SDEDM_FSM_DATAMODE)) {
-		pr_err("%s: previous command (%d) not complete (EDM %x)\n",
-		       mmc_hostname(host->mmc),
-		       bcm2835_sdhost_read(host, SDCMD) & SDCMD_CMD_MASK,
-		       edm);
 		log_event("REQ!", (u32)mrq, edm);
-		log_dump();
-		bcm2835_sdhost_dumpregs(host);
+		if (host->debug) {
+			pr_warn("%s: previous command (%d) not complete (EDM %x)\n",
+			       mmc_hostname(host->mmc),
+			       bcm2835_sdhost_read(host, SDCMD) & SDCMD_CMD_MASK,
+			       edm);
+			log_dump();
+			bcm2835_sdhost_dumpregs(host);
+		}
 		mrq->cmd->error = -EILSEQ;
 		tasklet_schedule(&host->finish_tasklet);
 		mmiowb();
@@ -1711,11 +1744,6 @@ static void bcm2835_sdhost_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
 
 	log_event("IOS<", ios->clock, 0);
 
-	if (!ios->clock || ios->clock != host->clock) {
-		bcm2835_sdhost_set_clock(host, ios->clock);
-		host->clock = ios->clock;
-	}
-
 	/* set bus width */
 	host->hcfg &= ~SDHCFG_WIDE_EXT_BUS;
 	if (ios->bus_width == MMC_BUS_WIDTH_4)
@@ -1731,6 +1759,9 @@ static void bcm2835_sdhost_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
 	mmiowb();
 
 	spin_unlock_irqrestore(&host->lock, flags);
+
+	if (!ios->clock || ios->clock != host->clock)
+		bcm2835_sdhost_set_clock(host, ios->clock);
 }
 
 static struct mmc_host_ops bcm2835_sdhost_ops = {
@@ -1794,16 +1825,19 @@ static void bcm2835_sdhost_tasklet_finish(unsigned long param)
 	mrq = host->mrq;
 
 	/* Drop the overclock after any data corruption, or after any
-	   error overclocked */
+	 * error while overclocked. Ignore errors for status commands,
+	 * as they are likely when a card is ejected. */
 	if (host->overclock) {
-		if ((mrq->cmd && mrq->cmd->error) ||
+		if ((mrq->cmd && mrq->cmd->error &&
+		     (mrq->cmd->opcode != MMC_SEND_STATUS)) ||
 		    (mrq->data && mrq->data->error) ||
-		    (mrq->stop && mrq->stop->error)) {
+		    (mrq->stop && mrq->stop->error) ||
+		    (mrq->sbc && mrq->sbc->error)) {
 			host->overclock_50--;
 			pr_warn("%s: reducing overclock due to errors\n",
 				mmc_hostname(host->mmc));
-			bcm2835_sdhost_set_clock(host,50*MHZ);
-			mrq->cmd->error = -EILSEQ;
+			host->reset_clock = 1;
+			mrq->cmd->error = -ETIMEDOUT;
 			mrq->cmd->retries = 1;
 		}
 	}
@@ -1828,6 +1862,21 @@ static void bcm2835_sdhost_tasklet_finish(unsigned long param)
 			       mmc_hostname(host->mmc), err);
 	}
 
+	/* The SDHOST block doesn't report any errors for a disconnected
+	   interface. All cards and SDIO devices should report some supported
+	   voltage range, so a zero response to SEND_OP_COND, IO_SEND_OP_COND
+	   or APP_SEND_OP_COND can be treated as an error. */
+	if (((mrq->cmd->opcode == MMC_SEND_OP_COND) ||
+	     (mrq->cmd->opcode == SD_IO_SEND_OP_COND) ||
+	     (mrq->cmd->opcode == SD_APP_OP_COND)) &&
+	    (mrq->cmd->error == 0) &&
+	    (mrq->cmd->resp[0] == 0)) {
+		mrq->cmd->error = -ETIMEDOUT;
+		if (host->debug)
+			pr_info("%s: faking timeout due to zero OCR\n",
+				mmc_hostname(host->mmc));
+	}
+
 	mmc_request_done(host->mmc, mrq);
 	log_event("TSK>", (u32)mrq, 0);
 }
@@ -1841,8 +1890,6 @@ int bcm2835_sdhost_add_host(struct bcm2835_host *host)
 
 	mmc = host->mmc;
 
-	bcm2835_sdhost_reset_internal(host);
-
 	mmc->f_max = host->max_clk;
 	mmc->f_min = host->max_clk / SDCDIV_MAX_CDIV;
 
@@ -1874,7 +1921,7 @@ int bcm2835_sdhost_add_host(struct bcm2835_host *host)
 
 			cfg.direction = DMA_MEM_TO_DEV;
 			cfg.src_addr = 0;
-			cfg.dst_addr = host->phys_addr + SDDATA;
+			cfg.dst_addr = host->bus_addr + SDDATA;
 
 			ret = dmaengine_slave_config(host->dma_chan_rxtx, &cfg);
 
@@ -1882,7 +1929,7 @@ int bcm2835_sdhost_add_host(struct bcm2835_host *host)
 				host->dma_cfg_tx = cfg;
 
 				cfg.direction = DMA_DEV_TO_MEM;
-				cfg.src_addr = host->phys_addr + SDDATA;
+				cfg.src_addr = host->bus_addr + SDDATA;
 				cfg.dst_addr = 0;
 
 				ret = dmaengine_slave_config(host->dma_chan_rxtx, &cfg);
@@ -1959,10 +2006,11 @@ static int bcm2835_sdhost_probe(struct platform_device *pdev)
 	struct resource *iomem;
 	struct bcm2835_host *host;
 	struct mmc_host *mmc;
+	const __be32 *addr;
+	u32 msg[3];
 	int ret;
 
 	pr_debug("bcm2835_sdhost_probe\n");
-	log_init();
 	mmc = mmc_alloc_host(sizeof(*host), dev);
 	if (!mmc)
 		return -ENOMEM;
@@ -1970,7 +2018,6 @@ static int bcm2835_sdhost_probe(struct platform_device *pdev)
 	mmc->ops = &bcm2835_sdhost_ops;
 	host = mmc_priv(mmc);
 	host->mmc = mmc;
-	host->cmd_quick_poll_retries = 0;
 	host->pio_timeout = msecs_to_jiffies(500);
 	host->pio_limit = 1;
 	host->max_delay = 1; /* Warn if over 1ms */
@@ -1984,11 +2031,17 @@ static int bcm2835_sdhost_probe(struct platform_device *pdev)
 		goto err;
 	}
 
-	host->phys_addr = iomem->start + BCM2835_VCMMU_SHIFT;
-	pr_debug(" - ioaddr %lx, iomem->start %lx, phys_addr %lx\n",
+	addr = of_get_address(node, 0, NULL, NULL);
+	if (!addr) {
+		dev_err(dev, "could not get DMA-register address\n");
+		return -ENODEV;
+	}
+	host->bus_addr = be32_to_cpup(addr);
+	log_init(iomem->start - host->bus_addr);
+	pr_debug(" - ioaddr %lx, iomem->start %lx, bus_addr %lx\n",
 		 (unsigned long)host->ioaddr,
 		 (unsigned long)iomem->start,
-		 (unsigned long)host->phys_addr);
+		 (unsigned long)host->bus_addr);
 
 	if (node) {
 		/* Read any custom properties */
@@ -1997,7 +2050,7 @@ static int bcm2835_sdhost_probe(struct platform_device *pdev)
 				     &host->delay_after_stop);
 		of_property_read_u32(node,
 				     "brcm,overclock-50",
-				     &host->overclock_50);
+				     &host->user_overclock_50);
 		of_property_read_u32(node,
 				     "brcm,pio-limit",
 				     &host->pio_limit);
@@ -2059,6 +2112,16 @@ static int bcm2835_sdhost_probe(struct platform_device *pdev)
 	else
 		mmc->caps |= MMC_CAP_4_BIT_DATA;
 
+	msg[0] = 0;
+	msg[1] = ~0;
+	msg[2] = ~0;
+
+	rpi_firmware_property(rpi_firmware_get(NULL),
+			      RPI_FIRMWARE_SET_SDHOST_CLOCK,
+			      &msg, sizeof(msg));
+
+	host->firmware_sets_cdiv = (msg[1] != ~0);
+
 	ret = bcm2835_sdhost_add_host(host);
 	if (ret)
 		goto err;
